# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import email as pyemail

from arjuna.tpi.log import log_info
from arjuna.core.error import EmailBoxNotConnected, EmailNotReceived
from arjuna.interact.email.conditions import MailBoxConditions

from .email import Email, Emails

class MailBox:
    '''
    Represents a named mailbox like Inbox in the email server for a user account.

    Keyword Arguments:
        server: Arjuna's Email Server object.
        name: Name of the mailbox. Default is Inbox.
        readonly: If True, messages can bot be deleted in the mailbox. Default is False.
    '''

    def __init__(self, *, server, name="INBOX", readonly=False):
        self.__server = server
        self.__name = name
        self.__readonly = readonly
        self.__count = None
        self.__saved_count = None

        self.save_state()

    def _select(self):
        status, data = self._server._imap.select(self.name, readonly=self.readonly)
        if status != "OK":
            raise EmailBoxNotConnected(self._server, self.name, data[0].decode())

    def _search(self, sender=None, subject=None, content=None, max=5, latest=True, consider_saved_state=True):
        self._select()

        data = None
        status, data = self._server._imap.search(None, 'ALL')
        if status != "OK":
            raise EmailNotReceived(self._server, self.name, data[0].decode())
        elif not data:
            raise EmailNotReceived(self._server, self.name, "No emails found.")

        email_numbers = data[0].split()
        new_email_count = len(email_numbers)

        if consider_saved_state:
            if new_email_count <= self.__saved_count:
                raise EmailNotReceived(self._server, self.name, "No *new* emails found since last saved state.")
        
        if consider_saved_state:
            new_email_count = new_email_count - self.__saved_count
        if new_email_count < max:
            max = new_email_count
        track = 0
        out_list = []
        if latest:
            email_numbers = reversed(email_numbers)
        for num in email_numbers:
            if track == max:
                break
            email = self.email_at(num.decode())

            if sender and sender.lower() not in email.sender.lower():
                continue

            if subject and subject.lower() not in email.subject.lower():
                continue

            if content:
                found = False
                for c in email.contents:
                    if content in str(c):
                        found = True
                        break
                if not found:
                    continue

            out_list.append(email)
            track += 1

        if out_list:
            return Emails(out_list)

        raise EmailNotReceived(self._server, self.name, "No *new* emails found since last saved state.", sender=sender, subject=subject, content=content)

    def __load(self):
        from arjuna import log_debug
        log_debug("Selecting mailbox: {}".format(self.name))
        if self._server._ready:
            MailBoxConditions(self).Select().wait(max_wait=30)
            _, data = self._server._imap.search(None, 'ALL')
            self.__count = sum(1 for num in data[0].split())
            log_debug("Selected mailbox: {}".format(self.name))
        else:
            raise Exception("A mailbox can be selected only with active connection to IMAP server.")

    def new_emails(self, *, sender=None, subject=None, content=None, max=5, max_wait=None):
        '''
        New emails received since the previous saved state (saved using save_state call), as per provided optional filters.

        Keyword Arguments:
            sender: Filter with Email address from which email is received.
            subject: Filter with Case-insensitive sub-string in subject line.
            content: Filter with Case-insensitive sub-string in content.
            max: Maximum number of emails to be returned. Default is 5.
            max_wait: Maximum time in seconds to re-select mailbox or availablity of emails in the mailbox.

        Note:
            The wait is not a static wait and comes into play if no email meets the filtering criteria.
            When the logic finds the first matching email, it tries to find more, else the wait loop is exited.
        '''
        from arjuna import C
        if max_wait is None:
            max_wait = C("emailauto.max.wait")
        return MailBoxConditions(self).ReceiveEmail(max=max, sender=sender, subject=subject, content=content, latest=True).wait(max_wait=max_wait)

    def latest(self, *, sender=None, subject=None, content=None, max_wait=None):
        '''
        Latest email received since the previous saved state (saved using save_state call), as per provided optional filters.

        This is a shortcut method, and works same as providing max=1 to new_emails.

        Keyword Arguments:
            sender: Filter with Email address from which email is received.
            subject: Filter with Case-insensitive sub-string in subject line.
            content: Filter with Case-insensitive sub-string in content.
            max_wait: Maximum time in seconds to re-select mailbox or availablity of emails in the mailbox.
        '''
        return self.new_emails(max=1, sender=sender, subject=subject, content=content, max_wait=max_wait)[0] 

    def emails(self, *, sender=None, subject=None, content=None, max=5, latest=True):
        '''
        Get Current Emails as per provided optional filters.

        Keyword Arguments:
            sender: Filter with Email address from which email is received.
            subject: Filter with Case-insensitive sub-string in subject line.
            content: Filter with Case-insensitive sub-string in content.
            max: Maximum number of emails to be returned. Default is 5.
            latest: If True, latest emails are considered. Default is True.

        Note:
            If no email is found in current emails that meets the filtering criteria, an automatic wait loop triggers and this method works just like new_emails method in such a context.
            When the logic finds the first matching email, it tries to find more, else the wait loop is exited.
        '''
        from arjuna import C
        max_wait = C("emailauto.max.wait")
        MailBoxConditions(self).Select().wait(max_wait=max_wait)
        return MailBoxConditions(self).ReceiveEmail(max=max, sender=sender, subject=subject, content=content, consider_saved_state=False).wait(max_wait=max_wait)

    @property
    def _server(self):
        return self.__server

    @property
    def name(self):
        '''
        Name of this mailbox.
        '''
        return self.__name

    @property
    def readonly(self):
        '''
        Is it a readonly mailbox selection?
        '''
        return self.__readonly

    @property
    def count(self):
        '''
        Current number of emails present in the mailbox.
        '''
        self.__load()
        return self.__count

    def save_state(self):
        '''
        Save current state information (like email count) before triggering a wait event in subsequent call.
        '''
        self.__load()
        self.__saved_count = self.__count

    def _force_state(self, count):
        '''
        Meant for module testing.
        '''
        self.__saved_count = count

    def email_at(self, ordinal, _reload=True):
        '''
        Email at an ordinal/position - human counted, 1 is the oldest email position.
        '''
        from arjuna import log_debug
        log_debug("Reading email at ordinal >{}< from mailbox {} at email server >{}<".format(
            ordinal, self.name, self._server.host
        ))
        if _reload:
            self.__load()
        _, data = self._server._imap.fetch(str(ordinal), '(RFC822)')
        return Email(pyemail.message_from_string(data[0][1].decode()))


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Very useful information on emails in Python: https://coderzcolumn.com/tutorials/python/email-how-to-represent-an-email-message-in-python


from .server import ImapServer


class EmailServer:

    @classmethod
    def imap(cls, *, host=None, port=None, user=None, password=None, use_ssl=None):
        '''
        Create an IMAP Server supporting IMAP4 protocol for reading emails.

        Keyword Arguments:
            host: IP or domain name of Email IMAP Server. Default is localhost. Default is configurable with emailauto.imap.host option.
            port: Port of the IMAP Server. Default is 143 for non-SSL Mode and 993 for SSL-Mode. Default is configurable with emailauto.imap.port option.
            user: Email address to open mailbox. Default is configurable with emailauto.user option.
            password: Password corresponding to the email account. Default is configurable with emailauto.password option.
            use_ssl: Should IMAPS (SSL-mode) be used? Default is True.
        '''
        return ImapServer(host=host, port=port, user=user, password=password, use_ssl=use_ssl)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.interact.gui.auto.finder.wmd import SimpleGuiWidgetMetaData
from arjuna.tpi.guiauto.meta.locator import GuiWidgetDefinition

from arjuna.tpi.guiauto.source.element import GuiElementSource
from arjuna.tpi.tracker import track
from arjuna.tpi.guiauto.model.gns import GNS

@track("debug")
class GuiFrame:
    '''
        Represents an IFrame the Gui.

        Not meant to be directly created. It is created using calls from **Gui** object or **GuiNamespace** object of **Gui**.

        Arguments:
            gui: **Gui** object containing this GuiDropDown.
            wmd: **GuiElementMetaData** object for this GuiDropDown.

        Keyword Arguments:
            parent: **GuiElement** in case it is found inside a **GuiElement**. Default is the **Gui** object.
    '''

    def __init__(self, gui, wmd, parent=None):
        self.__gui = gui
        self.__wmd = wmd
        self.__automator = gui._automator
        self.__frame_finder = parent and parent or gui
        from arjuna.tpi.guiauto.model.app import GuiApp
        from arjuna.tpi.guiauto.model.page import GuiPage
        from arjuna.tpi.guiauto.model.section import GuiSection

        # If Gui is an app, then it falls back to App's finding mechanism.
        # If Gui is an GuiPage, then it falls back to GuiPage's finding mechanism.
        # If Gui is an GuiSection, then it falls back to main GuiPage's finding mechanism.
        # This is because GuiSection might use nested element finding which will not work inside frame content.
        if isinstance(gui, GuiApp):
            self.__finder = gui.app
        elif isinstance(gui, GuiPage):
            self.__finder = gui
        elif isinstance(gui, GuiSection):
            self.__finder = gui.parent
        else:
            raise Exception("Unrecognized container GUI for Frame. Expecting App/Page/Section. Found: {}".format(gui))

        frame_element = self.__frame_finder._wmd_finder.element(wmd)
        self.__automator.dispatcher.switch_to_frame(frame_element.dispatcher)

        # Frame has new DOM. Hence it uses the GuiPage to do the finding and is dependent on GuiPage's GNS as well.
        self.__gns = GNS(self.__finder, self.__finder._gui_def)

    def exit(self):
        self.__automator.browser.switch_to_dom_root()

    @property
    def gns(self):
        '''
            Gui Namespace (GNS) object for this GuiFrame.
        '''
        return self.__gns

    @property
    def _wmd(self):
        return self.__wmd

    @property
    def gui(self) -> 'Gui':
        '''
            **Gui** object containing this GuiFrame.
        '''
        return self.__gui

    @property
    def source(self) -> GuiElementSource:
        '''
            **GuiSource** for this GuiDropDown (source of root element).
        '''
        return self._wrapped_main_element.source

    def locate(self, locator):
        '''
           Locate a GuiWidget.

           Arguments:
            locator: `GuiWidgetDefinition` object.

            Returns:
                An object of type `GuiWidget`. Exact object type depends on the value of **type** attribute in `GuiWidgetDefinition`. 
        '''
        return self.__finder.locate(locator)

    def element(self, *, fargs=None, **kwargs) -> 'GuiElement':
        '''
            Locate a `GuiElement`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetDefinition`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetDefinition`

            Returns:
                `GuiElement` object.
        '''
        return self.__finder.element(*fargs, **kwargs)

    def multi_element(self, fargs=None, **kwargs) -> 'GuiMultiElement':
        '''
            Locate a `GuiMultiElement`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetDefinition`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetDefinition`

            Returns:
                `GuiMultiElement` object.
        '''
        return self.__finder.multi_element(*fargs, **kwargs)

    def dropdown(self, fargs=None, **kwargs) -> 'GuiDropDown':
        '''
            Locate a `GuiDropDown`.

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetDefinition`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetDefinition`

            Returns:
                `GuiDropDown` object.
        '''
        return self.__finder.dropdown(*fargs, **kwargs)

    def radio_group(self, fargs=None, **kwargs) -> 'GuiRadioGroup':
        '''
            Locate a `GuiRadioGroup`

            Keyword Arguments:
                fargs: A dictionary of key-value pairs for formatting the `GuiWidgetDefinition`. Use **.format(kwargs).wait_until_absent** for more Pythonic code when formatting.
                **kwargs: Arbitrary key-value pairs used to construct a `GuiWidgetDefinition`

            Returns:
                `GuiRadioGroup` object
        '''
        return self.__finder.radio_group(*fargs, **kwargs)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from email.header import decode_header

from arjuna.tpi.log import log_info


class Email:
    '''
    Represents an Email.

    Keyword Arguments:
        pyemail: Python's builtin email object.
    '''

    def __init__(self, pyemail):
        self.__email = pyemail
        self.__metadata = {}
        for k in self.__email.keys():
            value, encoding = decode_header(self.__email.get(k))[0]
            if isinstance(value, bytes):
                self.__metadata[k] = value.decode(encoding)
            else:
                self.__metadata[k] = value
        self.__contents = []
        self.__load_content()

    def __str__(self):
        return str(self.__email)

    def get_header(self, name):
        '''
        Get value for an email-header.
        '''
        return self.__metadata[name]

    @property
    def sender(self):
        '''
        Sender email-address (From)
        '''
        return self.get_header("From")

    @property
    def recipient(self):
        '''
        Recipient email-address (To)
        '''
        return self.get_header("To")

    @property
    def date(self):
        '''
        Date of receipt of this email.
        '''
        return self.get_header("Date")

    @property
    def subject(self):
        '''
        Subject line of this email.
        '''
        return self.get_header("Subject")

    @property
    def content_type(self):
        '''
        Content-Type of this email.
        '''
        return self.get_header("Content-Type")

    @property
    def is_multipart(self):
        '''
        Is it a multi-part email?
        '''
        return self.__email.is_multipart()

    def _load_part_content(self, part):
        # Code heavily taken from https://www.thepythoncode.com/article/reading-emails-in-python
        content_type = part.get_content_type()
        content_disposition = str(part.get("Content-Disposition"))
        try:
            # Email body
            body = part.get_payload(decode=True).decode()
        except:
            return
        if "attachment" not in content_disposition:
            if content_type == "multipart/alternative":
                pass
            elif content_type =="text/plain":
                from arjuna.tpi.parser.text import Text
                self.__contents.append(Text(body))
            elif content_type =="text/html":
                from arjuna.tpi.parser.html import Html
                self.__contents.append(Html.from_str(body))
            else:
                print("Unsupported email content type.", content_type)
                # Unsupported content type
                pass  
        else:
            pass
            # Currently Arjuna does not handle email attachments

    def _walk(self, email):
        for index, part in enumerate(email.walk()):
            self._load_part_content(part)

    def __load_content(self):
        if self.is_multipart:
            self._walk(self.__email)
        else:
            self._load_part_content(self.__email)

    @property
    def contents(self):
        '''
        Contents of this email as a list of objects representing different parts. Currently attachments are not processed.
        '''
        return self.__contents

    def find_links(self, *, unique=True, contain=""):
        '''
        All hyperlinks in the email content.

        Keyword Arguments:
            unique: Only unique hyperlinks are included in result.
            contain: Only hyperlinks that contain the provided substring are included in the result.
        '''
        outlist = []
        for content in self.contents:
            out = content.find_links(unique=unique, contain=contain)
            if out:
                outlist.extend(out)

        if unique:
            return tuple(set(outlist))
        else:
            return tuple(outlist) 

    def find_link(self, contains):
        '''
        Returns first hyperlink that contains the provided URI part as a sub-string.

        Arguments:
            contains: A sub-string which should be present in the hyperlink.
        '''
        links = self.find_links(contain=contains)
        if links:
            return links[0]
        else:
            raise Exception("No Hyperlink was found with sub-string >{}<".format(contains))


    @property
    def links(self):
        '''
        All hyperlinks in the email content.
        '''
        return self.find_links(unique=False)

    @property
    def unique_links(self):
        '''
        All unique hyperlinks in the email content.
        '''
        return self.find_links()


class Emails:
    '''
    Represents a sequence of one or more Arjuna's Email objects.

    Keyword Arguments:
        email_list: Sequence of Arjuna's email objects.

    Note:
        You can loop over this object like a Python-list.

        .. code-block:: python

            for email in emails:
                print(email)
    '''


    def __init__(self, email_list):
        self.__emails = email_list

    @property
    def count(self):
        '''
        Number of emails in this list.
        '''
        return len(self.__emails)

    def __len__(self):
        return self.count

    def __getitem__(self, index):
        return self.__emails[index]

    def __iter__(self):
        return iter(self.__emails)


# This file is a part of Arjuna
# Copyright 2015-2021 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imaplib

from .mailbox import MailBox


class ImapServer:
    '''
    IMAP Server supporting IMAP4 protocol for reading emails.

    Keyword Arguments:
        host: IP or domain name of Email IMAP Server. Default is localhost. Default is configurable with emailauto.imap.host option.
        port: Port of the IMAP Server. Default is 143 for non-SSL Mode and 993 for SSL-Mode. Default is configurable with emailauto.imap.port option.
        user: Email address to open mailbox. Default is configurable with emailauto.user option.
        password: Password corresponding to the email account. Default is configurable with emailauto.password option.
        use_ssl: Should IMAPS (SSL-mode) be used? Default is True.
    '''

    def __init__(self, *, host=None, port=None, user=None, password=None, use_ssl=None):
        from arjuna import C
        cport = C("emailauto.imap.port")
        self.__host = host is None and C("emailauto.imap.host") or host
        if use_ssl is None:
            self.__use_ssl = C("emailauto.imap.usessl")
        else:
            self.__use_ssl = use_ssl
        if port is None:
            if cport != "not_set":
                self.__port = cport
            else:
                self.__port = self.__use_ssl and 993 or 143
        self.__user = user and user or C("emailauto.user")
        if self.__user == "not_set":
            raise Exception("One must provide IMAP server user in the constructor or in the reference configuration value for emailauto.user option.")

        self.__password = password and password or C("emailauto.password")
        if self.__password == "not_set":
            raise Exception("One must provide IMAP server password in the constructor or in the reference configuration value for emailauto.password option.")

        self.__imap = None
        self.__connected = False
        self.__loggedin = False
        self.__mailbox = None
        self.__ready = False

        self.connect()

    def connect(self, reconnect=False):
        '''
        Connect to the server and login using the provided account details.

        Keyword Arguments:
            reconnect: If already connected, force a reconnect. Default is False.
        '''
        if self.__ready:
            if not reconnect:
                return
            else:
                self.quit()
        from arjuna import log_info
        log_info("Connecting to IMAP server >{}< at port >{}< (Using SSL: {}) for user: >{}<.".format(
            self.host, self.port, self.uses_ssl, self.user
        ))
        if not self.__connected:
            if self.uses_ssl:
                self.__imap = imaplib.IMAP4_SSL(self.host, self.port)
            else:
                self.__imap = imaplib.IMAP4(self.host, self.port)
            self.__connected = True

        if not self.__loggedin:
            self._imap.login(self.user, self.password)
            self.__loggedin = True

        # By default connects to Inbox.
        self.__ready = True
        log_info("Connected to IMAP Server >{}<.".format(self.host))

    def quit(self):
        '''
        Disconnect from the server and logout.
        '''
        from arjuna import log_info
        log_info("Disconnecting from IMAP Server >{}<.".format(self.host))
        if self.__ready:
            # Select should be done atleast once before closing. So this line takes care of that.
            self.__imap.select("INBOX")
            self._imap.close()
            self.__msg_count = None
            if self.__loggedin:
                self._imap.logout()
                self.__loggedin = False
            log_info("Disconnected from IMAP Server >{}<.".format(self.host))
            self.__connected = False
            return
        log_info("Quit called without active connection to IMAP Server >{}<. Ignored.".format(self.host))

    def get_mailbox(self, name="INBOX", *, readonly=False):
        '''
        Get a MailBox object.

        Arguments:
            name: Name of the mailbox. Default is Inbox.

        Keyword Arguments:
            readonly: If True, messages can bot be deleted in the mailbox. Default is False.
        '''
        return MailBox(server=self, name=name, readonly=readonly)

    @property
    def host(self):
        '''
        Domain/IP Address of the server.
        '''
        return self.__host

    @property
    def uses_ssl(self):
        '''
        Does it use SSL?
        '''
        return self.__use_ssl

    @property
    def port(self):
        '''
        Network port of the server.
        '''
        return self.__port

    @property
    def user(self):
        '''
        Email account to open mailbox.
        '''
        return self.__user

    @property
    def password(self):
        '''
        Email account password to open mailbox.
        '''
        return self.__password

    @property
    def _imap(self):
        return self.__imap

    @property
    def _ready(self):
        return self.__ready

    @property
    def _mailbox(self):
        return self.__mailbox